.386p
.MODEL FLAT

	INCLUDE "audio/apaula.i"
	INCLUDE "hardware/avideo.i"

SCOPES_X	EQU 128
SCOPES_Y	EQU 55
SCOPE_W		EQU 40
SCOPE_H		EQU 33
SCOPE_CENTER	EQU 16
SCOPE_PADDING	EQU 8
SCOPE_NUM	EQU 4

	; ui.c
	EXTERN _ScopeTable:DWORD

	; audio.c
	EXTERN _audioOpen:BYTE

	PUBLIC _Scopes_GenerateTable
	PUBLIC _Scopes_ResetStates
	PUBLIC _Scopes_Draw

.CODE

	ALIGN 4
_Scopes_GenerateTable PROC NEAR
	pushad
	; --------------------------
	mov edi,_ScopeTable
	mov edx,((32+1)*256)-1
	lea edi,[edi+edx*2]		; start at end of LUT
	mov cl,8
	mov bl,SCREEN_W/8
	std				; write in reverse
	; --------------------------
GTLoop:	mov eax,edx			; cl = -128..127 (sample data)
	shr eax,cl			; al = volume (0..32)
	neg al				; negate volume (-32..0)
	imul dl				; ax = dl * volume
	sar eax,cl			; al = -15..16
	add al,16			; al = 0..32
	mul bl				; ax = al * (SCREEN_W/8)
	stosw
	dec edx
	jns GTLoop
	; --------------------------
	cld
	popad
	retn
_Scopes_GenerateTable ENDP

	ALIGN 4
_Scopes_ResetStates PROC NEAR
	push eax
	xor eax,eax
	mov dword ptr ScopeDrawn,eax
	mov dword ptr EmptyScopeDrawn,eax
	pop eax
	retn
_Scopes_ResetStates ENDP


SCLEAR_M MACRO pos
	mov ax,[esi-(pos*2)]
	mov [edi+eax],dl
ENDM

	; Input:
	;  EAX = X position
	;  EBX = Paula voice struct for current channel
	;  ECX = Scope number
	ALIGN 4
drawOneScope PROC NEAR
	pushad
	; --------------------------
	push ecx
	shr eax,3
	lea edi,[VGA_FRAMEBUFFER+SCOPES_Y*(SCREEN_W/8)+eax]
	; --------------------------
	mov ebp,LastValOffs[ecx*4]
	mov LastValPtr,ebp
	; --------------------------
	cmp byte ptr _audioOpen,0
	je Inactive
	; --------------------------
	cmp dword ptr v_dmaActive[ebx],0
	je Inactive
	; --------------------------
	;cmp byte ptr v_volume[ebx],0
	;je Inactive
	; --------------------------
	; if looping 2-byte loop, we're inactive
	cmp dword ptr v_length[ebx],2
	ja V0
	cmp dword ptr v_newLength[ebx],2
	jbe Inactive
	; --------------------------
V0:	cmp byte ptr ScopeDrawn[ecx],0
	je NoWipe
	; --------------------------
	; erase last scope VRAM pixels
	; --------------------------
	mov esi,ebp
	mov ebp,8*2
	add esi,(SCOPE_W*2)-2
	xor eax,eax ; upper bytes must be cleared
	xor edx,edx
	; --------------------------
	mov ecx,SCOPE_W/8
DLoop0:	SCLEAR_M 0
	SCLEAR_M 1
	SCLEAR_M 2
	SCLEAR_M 3
	SCLEAR_M 4
	SCLEAR_M 5
	SCLEAR_M 6
	SCLEAR_M 7
	sub esi,ebp
	inc edi
	dec ecx
	jnz DLoop0
	sub edi,SCOPE_W/8
NoWipe:	; --------------------------
	; set VGA latch operation to OR
	mov dx,3CEh
	mov ax,1003h
	out dx,ax
	; --------------------------
	; setup volume table pointer
	movzx edx,byte ptr v_volume[ebx]
	shr edx,1			; 0..64 -> 0..32
	shl edx,8+1			; ebx *= 256*2
	mov eax,_ScopeTable		; (pointer is a multiple of 512)
	add eax,edx
	shr eax,1
	; --------------------------
	mov esi,v_dataPtr[ebx]
	mov edx,LastValPtr
	; --------------------------
	mov ebp,v_dataEnd[ebx]
	mov dword ptr [CmpVal+2],ebp
	jmp DS_C			; invalidate cache since we modified code
DS_C:	; --------------------------
	mov ebp,v_newDataPtr[ebx]
	mov NewDataPtr,ebp
	mov ebp,v_newLength[ebx]
	mov NewLength,ebp
	xor ebp,ebp
	; --------------------------
	; write sample points to VRAM
	; --------------------------
	mov ebx,SCOPE_W-1
	mov cl,10000000b		; current bitplane bit
	cld
DLoop1:	lodsb
	mov bp,[eax+eax]
	mov [edx+ebx*2],bp		; store sample Y position in "last y" buffer
	or [edi+ebp],cl			; slow read-write-modify VGA access (needed)
	ror cl,1
	adc edi,0
CmpVal:	cmp esi,123456789		; end of sample reached?
	jae SampleSwap			; time to update location and location end
	dec ebx
	jns DLoop1
	; --------------------------
	jmp DSDone

SampleSwap:
	mov esi,NewDataPtr
	mov ebp,NewLength
	add ebp,esi			; new sample end address
	mov dword ptr [CmpVal+2],ebp	; modify code (cache gets invalidated on DLoop1 branch)
	xor ebp,ebp
	dec ebx
	jns DLoop1
	; --------------------------

DSDone:	; --------------------------
	; set VGA latch operation back to normal
	mov dx,3CEh
	mov ax,0003h
	out dx,ax
	; --------------------------
DSEnd:	pop ecx
	mov byte ptr ScopeDrawn[ecx],1
	mov byte ptr EmptyScopeDrawn[ecx],0
	; --------------------------
	popad
	retn

Inactive:
	; --------------------------
	; voice is not active
	; --------------------------
	pop ecx

	cmp byte ptr EmptyScopeDrawn[ecx],1
	je DSEnd2 ; we have already drawn the empty scope

	; clear "last pixels" buffer
	push edi
	push ecx
	mov edi,LastValPtr
	mov ecx,SCOPE_W/2
	mov eax,((SCOPE_CENTER*(SCREEN_W/8))*65536)+(SCOPE_CENTER*(SCREEN_W/8))
	cld
	rep stosd
	pop ecx
	pop edi

	; erase scope background (once until voice becomes active)
	xor eax,eax
	mov ebx,SCOPE_H/11
	mov edx,(SCREEN_W/8)*4
	mov ebp,(SCREEN_W/8)*3
DLoop2:	mov [edi+((SCREEN_W/8)*0)+0],eax
	mov [edi+((SCREEN_W/8)*0)+4],al
	mov [edi+((SCREEN_W/8)*1)+0],eax
	mov [edi+((SCREEN_W/8)*1)+4],al
	mov [edi+((SCREEN_W/8)*2)+0],eax
	mov [edi+((SCREEN_W/8)*2)+4],al
	mov [edi+((SCREEN_W/8)*3)+0],eax
	mov [edi+((SCREEN_W/8)*3)+4],al
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)+0],eax
	mov [edi+((SCREEN_W/8)*0)+4],al
	mov [edi+((SCREEN_W/8)*1)+0],eax
	mov [edi+((SCREEN_W/8)*1)+4],al
	mov [edi+((SCREEN_W/8)*2)+0],eax
	mov [edi+((SCREEN_W/8)*2)+4],al
	mov [edi+((SCREEN_W/8)*3)+0],eax
	mov [edi+((SCREEN_W/8)*3)+4],al
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)+0],eax
	mov [edi+((SCREEN_W/8)*0)+4],al
	mov [edi+((SCREEN_W/8)*1)+0],eax
	mov [edi+((SCREEN_W/8)*1)+4],al
	mov [edi+((SCREEN_W/8)*2)+0],eax
	mov [edi+((SCREEN_W/8)*2)+4],al
	add edi,ebp
	dec ebx
	jnz DLoop2

	; draw center line
	sub edi,((SCOPE_H/2)+1)*(SCREEN_W/8)
	dec eax ; 0FFFFFFFFh
	mov [edi+0],eax
	mov [edi+4],al

	mov byte ptr ScopeDrawn[ecx],1
	mov byte ptr EmptyScopeDrawn[ecx],1	; prevent rewrite every frame
DSEnd2:
	popad
	retn
drawOneScope ENDP

_Scopes_Draw PROC NEAR
	push eax
	push ebx
	push ecx
	; -----------------------------
	mov dx,3C4h ; set correct bitplanes	
	mov ax,0402h	
	out dx,ax
	; -----------------------------
	mov eax,SCOPES_X
	mov ebx,_paulaOffset
	xor ecx,ecx
Loop0:	call drawOneScope
	add eax,SCOPE_W+SCOPE_PADDING
	add ebx,v_sizeof
	inc ecx
	cmp cl,SCOPE_NUM
	jne Loop0
	; -----------------------------
	pop ecx
	pop ebx
	pop eax
	retn
_Scopes_Draw ENDP

.DATA

	ALIGN 4
LastValPtr	dd 0
ScopeDrawn	db SCOPE_NUM dup (0)
EmptyScopeDrawn	db SCOPE_NUM dup (0)

	ALIGN 4
LastValOffs:
	dd LastVal1
	dd LastVal2
	dd LastVal3
	dd LastVal4

	ALIGN 4
LastVal1 dw SCOPE_W dup (?)
	ALIGN 4
LastVal2 dw SCOPE_W dup (?)
	ALIGN 4
LastVal3 dw SCOPE_W dup (?)
	ALIGN 4
LastVal4 dw SCOPE_W dup (?)

	ALIGN 4
NewDataPtr	dd 0
NewLength	dd 0

END
