#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <conio.h>
#include "drivers/sb.h"
#include "paula.h"
#include "replayer.h" // Replayer_Tick()
#include "mixer.h"
#include "../ui.h" // ui.chmuted
#include "../hardware/video.h" // dFrameRate
#include "audio.h"

static bool audioDetected;

// globals
bool output16Bit, audioOpen;
void *mixBuffer;
uint32_t samplesPerTickHi, samplesPerTickLo, sampleCounterHi, sampleCounterLo;
audio_t audio;

void Audio_SetSamplesPerTick(int64_t samples)
{
	samplesPerTickHi = samples >> 32;
	samplesPerTickLo = (uint32_t)samples;
}

void Audio_ClearSampleCounter(void)
{
	sampleCounterHi = sampleCounterLo = 0;
}

bool Audio_Detect(bool noSound)
{
	if (noSound)
	{
		audioDetected = false;
		return false;
	}
	
	if (SB_Detect())
	{
		audioDetected = true;
		return true;
	}
	else
	{
		printf("WARNING: Couldn't detect a compatible Sound Blaster card.\n");
		printf("Make sure the BLASTER variable is (properly) set.\n\n");
		printf("Press any key to continue with no sound...\n");
		getch();

		audioDetected = false;
		return false;
	}
}

bool Audio_Init(void)
{
	if (!audioDetected)
	{
		audio.dOutputFreq = VBLANK_HZ * 512; // extra precision needed
		return true; // "no audio" mode
	}
	
	if (!SB_Init())
		return false;

	mixBuffer = (int32_t *)calloc(AUDIO_BUF_SIZE, sizeof (int32_t));
	if (mixBuffer == NULL)
	{
		printf("ERROR: Out of memory!\n");
		return false; // mem gets free'd later
	}

	audioOpen = true;
	return true;
}

bool Audio_StartDMA(void)
{
	if (!audioDetected)
		return true; // "no audio" mode

	SB_StartDMA();
	audio.soundReady = true;
	return true;
}

void Audio_Close(void)
{
	if (audio.soundReady)
	{
		SB_Close();
		audio.soundReady = false;
	}

	if (mixBuffer != NULL)
	{
		free(mixBuffer);
		mixBuffer = NULL;
	}
}
