#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include "gfx/font.h"
#include "ui.h"
#include "pattern.h"

static uint8_t *drawCodeAreaOrigPtr, *drawCodeArea;
static int32_t *oldRowsOrigPtr;

static const char *noteNames[1+(12*3)+1] =
{
	"---",
	"C-1", "C#1", "D-1", "D#1", "E-1", "F-1", "F#1" ,"G-1", "G#1", "A-1", "A#1", "B-1",
	"C-2", "C#2", "D-2", "D#2", "E-2", "F-2", "F#2" ,"G-2", "G#2", "A-2", "A#2", "B-2",
	"C-3", "C#3", "D-3", "D#3", "E-3", "F-3", "F#3" ,"G-3", "G#3", "A-3", "A#3", "B-3",
	"???"
};

static const char HexTable[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

// globalized, for apattern.asm
uint32_t noteDrawRoutines[1+(12*3)+1], oneHexDrawRoutines[16], twoHexDrawRoutines[256];
uint32_t rowDrawRoutines[64];
uint32_t bigNoteDrawRoutines[1+(12*3)+1], bigOneHexDrawRoutines[16], bigTwoHexDrawRoutines[256];
uint32_t bigRowDrawRoutines[64];
int32_t *oldRows;
// ----------------------------

bool DrawCode_MakeCode(void)
{
	uint8_t ch1, ch2, ch3, *fontData;
	uint16_t index;
	uint32_t a, i, x, y;

	oldRowsOrigPtr = (int32_t *)malloc((VISIBLE_ROWS * 4 * 4 * sizeof (int32_t)) + 4);
	if (oldRowsOrigPtr == NULL)
		return false;
	oldRows = (int32_t *)(((uint32_t)oldRowsOrigPtr + 3) & ~3); // dword aligned	

	drawCodeAreaOrigPtr = (uint8_t *)malloc(48124 + 4); // total code size (check 'a' variable after code generation end)
	if (drawCodeAreaOrigPtr == NULL)
		return false;
	drawCodeArea = (uint8_t *)(((uint32_t)drawCodeAreaOrigPtr + 3) & ~3); // dword aligned	

	// create pattern drawing code (inspired by FT2 source code)

	fontData = Font_GetData();
	a = 0; // code address

	// row (2 characters, 0..63)
	for (i = 0; i < 64; i++)
	{
		rowDrawRoutines[i] = (uint32_t)&drawCodeArea[a];
		for (x = 0; x < 2; x++)
		{
			if (x == 0)
				ch1 = ('0' + (i / 10)) - ' ';
			else
				ch1 = ('0' + (i % 10)) - ' ';

			for (y = 0; y < FONT_H; y++)
			{
				uint8_t data = fontData[(ch1*6) + y];
				
				index = (y*(SCREEN_W / 8)) + x;
				if (index < 128)
				{
					// mov byte ptr [edi+XXh],YYh
					drawCodeArea[a+0] = 0xC6;
					drawCodeArea[a+1] = 0x47;
					drawCodeArea[a+2] = (uint8_t)index;
					drawCodeArea[a+3] = data;
					a += 4;
				}
				else
				{
					// mov byte ptr [edi+XXXXXXXXh],YYh
					drawCodeArea[a+0] = 0xC6;
					drawCodeArea[a+1] = 0x87;
					drawCodeArea[a+2] = index & 0xFF;
					drawCodeArea[a+3] = index >> 8;
					drawCodeArea[a+4] = 0x00;
					drawCodeArea[a+5] = 0x00;
					drawCodeArea[a+6] = data;
					a += 7;
				}
			}
		}

		drawCodeArea[a] = 0xC3; // retn
		a++;
	}

	// big row (2 characters, 0..63)
	for (i = 0; i < 64; i++)
	{
		bigRowDrawRoutines[i] = (uint32_t)&drawCodeArea[a];
		for (x = 0; x < 2; x++)
		{
			if (x == 0)
				ch1 = ('0' + (i / 10)) - ' ';
			else
				ch1 = ('0' + (i % 10)) - ' ';

			for (y = 0; y < FONT_H; y++)
			{
				uint8_t data = fontData[(ch1*6) + y];
				
				index = ((y<<1)*(SCREEN_W / 8)) + x;
				if (index < 128)
				{
					// mov byte ptr [edi+XXh],YYh
					drawCodeArea[a+0] = 0xC6;
					drawCodeArea[a+1] = 0x47;
					drawCodeArea[a+2] = (uint8_t)index;
					drawCodeArea[a+3] = data;
					a += 4;
				}
				else
				{
					// mov byte ptr [edi+XXXXXXXXh],YYh
					drawCodeArea[a+0] = 0xC6;
					drawCodeArea[a+1] = 0x87;
					drawCodeArea[a+2] = index & 0xFF;
					drawCodeArea[a+3] = index >> 8;
					drawCodeArea[a+4] = 0x00;
					drawCodeArea[a+5] = 0x00;
					drawCodeArea[a+6] = data;
					a += 7;
				}

				index += SCREEN_W / 8;
				if (index < 128)
				{
					// mov byte ptr [edi+XXh],YYh
					drawCodeArea[a+0] = 0xC6;
					drawCodeArea[a+1] = 0x47;
					drawCodeArea[a+2] = (uint8_t)index;
					drawCodeArea[a+3] = data;
					a += 4;
				}
				else
				{
					// mov byte ptr [edi+XXXXXXXXh],YYh
					drawCodeArea[a+0] = 0xC6;
					drawCodeArea[a+1] = 0x87;
					drawCodeArea[a+2] = index & 0xFF;
					drawCodeArea[a+3] = index >> 8;
					drawCodeArea[a+4] = 0x00;
					drawCodeArea[a+5] = 0x00;
					drawCodeArea[a+6] = data;
					a += 7;
				}
			}
		}

		drawCodeArea[a] = 0xC3; // retn
		a++;
	}

	// notes (3 characters, 0..36. 37 = illegal note)
	for (i = 0; i < 1+(12*3)+1; i++)
	{
		noteDrawRoutines[i] = (uint32_t)&drawCodeArea[a];

		ch1 = noteNames[i][0] - ' ';
		ch2 = noteNames[i][1] - ' ';
		ch3 = noteNames[i][2] - ' ';

		for (y = 0; y < FONT_H; y++)
		{
			uint8_t data1 = fontData[(ch1*6) + y];
			uint8_t data2 = fontData[(ch2*6) + y];
			uint8_t data3 = fontData[(ch3*6) + y];
			
			index = y*(SCREEN_W / 8);
			if (index < 128)
			{
				// mov word ptr [edi+XXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x47;
				drawCodeArea[a+3] = (uint8_t)index;
				drawCodeArea[a+4] = data1;
				drawCodeArea[a+5] = data2;
				a += 6;
			}
			else
			{
				// mov word ptr [edi+XXXXXXXXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x87;
				drawCodeArea[a+3] = index & 0xFF;
				drawCodeArea[a+4] = index >> 8;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = 0x00;
				drawCodeArea[a+7] = data1;
				drawCodeArea[a+8] = data2;
				a += 9;
			}

			index = (y*(SCREEN_W / 8)) + 2;
			if (index < 128)
			{
				// mov byte ptr [edi+XXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x47;
				drawCodeArea[a+2] = (uint8_t)index;
				drawCodeArea[a+3] = data3;
				a += 4;
			}
			else
			{
				// mov byte ptr [edi+XXXXXXXXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x87;
				drawCodeArea[a+2] = index & 0xFF;
				drawCodeArea[a+3] = index >> 8;
				drawCodeArea[a+4] = 0x00;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = data3;
				a += 7;
			}
		}

		drawCodeArea[a] = 0xC3; // retn
		a++;
	}

	// big notes (3 characters, 0..36. 37 = illegal note)
	for (i = 0; i < 1+(12*3)+1; i++)
	{
		bigNoteDrawRoutines[i] = (uint32_t)&drawCodeArea[a];

		ch1 = noteNames[i][0] - ' ';
		ch2 = noteNames[i][1] - ' ';
		ch3 = noteNames[i][2] - ' ';

		for (y = 0; y < FONT_H; y++)
		{
			uint8_t data1 = fontData[(ch1*6) + y];
			uint8_t data2 = fontData[(ch2*6) + y];
			uint8_t data3 = fontData[(ch3*6) + y];

			index = (y<<1)*(SCREEN_W / 8);
			if (index < 128)
			{
				// mov word ptr [edi+XXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x47;
				drawCodeArea[a+3] = (uint8_t)index;
				drawCodeArea[a+4] = data1;
				drawCodeArea[a+5] = data2;
				a += 6;
			}
			else
			{
				// mov word ptr [edi+XXXXXXXXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x87;
				drawCodeArea[a+3] = index & 0xFF;
				drawCodeArea[a+4] = index >> 8;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = 0x00;
				drawCodeArea[a+7] = data1;
				drawCodeArea[a+8] = data2;
				a += 9;
			}

			index = ((y<<1)*(SCREEN_W / 8)) + 2;
			if (index < 128)
			{
				// mov byte ptr [edi+XXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x47;
				drawCodeArea[a+2] = (uint8_t)index;
				drawCodeArea[a+3] = data3;
				a += 4;
			}
			else
			{
				// mov byte ptr [edi+XXXXXXXXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x87;
				drawCodeArea[a+2] = index & 0xFF;
				drawCodeArea[a+3] = index >> 8;
				drawCodeArea[a+4] = 0x00;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = data3;
				a += 7;
			}

			index = ((y<<1)+1)*(SCREEN_W / 8);
			if (index < 128)
			{
				// mov word ptr [edi+XXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x47;
				drawCodeArea[a+3] = (uint8_t)index;
				drawCodeArea[a+4] = data1;
				drawCodeArea[a+5] = data2;
				a += 6;
			}
			else
			{
				// mov word ptr [edi+XXXXXXXXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x87;
				drawCodeArea[a+3] = index & 0xFF;
				drawCodeArea[a+4] = index >> 8;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = 0x00;
				drawCodeArea[a+7] = data1;
				drawCodeArea[a+8] = data2;
				a += 9;
			}

			index = (((y<<1)+1)*(SCREEN_W / 8)) + 2;
			if (index < 128)
			{
				// mov byte ptr [edi+XXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x47;
				drawCodeArea[a+2] = (uint8_t)index;
				drawCodeArea[a+3] = data3;
				a += 4;
			}
			else
			{
				// mov byte ptr [edi+XXXXXXXXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x87;
				drawCodeArea[a+2] = index & 0xFF;
				drawCodeArea[a+3] = index >> 8;
				drawCodeArea[a+4] = 0x00;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = data3;
				a += 7;
			}
		}

		drawCodeArea[a] = 0xC3; // retn
		a++;
	}

	// two hex (2 characters, 0..FF)
	for (i = 0; i < 256; i++)
	{
		twoHexDrawRoutines[i] = (uint32_t)&drawCodeArea[a];

		ch1 = HexTable[i >> 4] - ' ';
		ch2 = HexTable[i & 0x0F] - ' ';
		for (y = 0; y < FONT_H; y++)
		{
			const uint16_t data1 = fontData[(ch1*6) + y];
			const uint16_t data2 = fontData[(ch2*6) + y];

			index = y*(SCREEN_W / 8);
			if (index < 128)
			{
				// mov word ptr [edi+XXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x47;
				drawCodeArea[a+3] = (uint8_t)index;
				drawCodeArea[a+4] = data1;
				drawCodeArea[a+5] = data2;
				a += 6;
			}
			else
			{
				// mov word ptr [edi+XXXXXXXXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x87;
				drawCodeArea[a+3] = index & 0xFF;
				drawCodeArea[a+4] = index >> 8;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = 0x00;
				drawCodeArea[a+7] = data1;
				drawCodeArea[a+8] = data2;
				a += 9;
			}
		}

		drawCodeArea[a] = 0xC3; // retn
		a++;
	}

	// big two hex (2 characters, 0..FF)
	for (i = 0; i < 256; i++)
	{
		bigTwoHexDrawRoutines[i] = (uint32_t)&drawCodeArea[a];

		ch1 = HexTable[i >> 4] - ' ';
		ch2 = HexTable[i & 0x0F] - ' ';
		for (y = 0; y < FONT_H; y++)
		{
			const uint16_t data1 = fontData[(ch1*6) + y];
			const uint16_t data2 = fontData[(ch2*6) + y];
			
			index = (y<<1)*(SCREEN_W / 8);
			if (index < 128)
			{
				// mov word ptr [edi+XXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x47;
				drawCodeArea[a+3] = (uint8_t)index;
				drawCodeArea[a+4] = data1;
				drawCodeArea[a+5] = data2;
				a += 6;
			}
			else
			{
				// mov word ptr [edi+XXXXXXXXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x87;
				drawCodeArea[a+3] = index & 0xFF;
				drawCodeArea[a+4] = index >> 8;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = 0x00;
				drawCodeArea[a+7] = data1;
				drawCodeArea[a+8] = data2;
				a += 9;
			}

			index = ((y<<1)+1)*(SCREEN_W / 8);
			if (index < 128)
			{
				// mov word ptr [edi+XXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x47;
				drawCodeArea[a+3] = (uint8_t)index;
				drawCodeArea[a+4] = data1;
				drawCodeArea[a+5] = data2;
				a += 6;
			}
			else
			{
				// mov word ptr [edi+XXXXXXXXh],YYYYh
				drawCodeArea[a+0] = 0x66;
				drawCodeArea[a+1] = 0xC7;
				drawCodeArea[a+2] = 0x87;
				drawCodeArea[a+3] = index & 0xFF;
				drawCodeArea[a+4] = index >> 8;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = 0x00;
				drawCodeArea[a+7] = data1;
				drawCodeArea[a+8] = data2;
				a += 9;
			}
		}

		drawCodeArea[a] = 0xC3; // retn
		a++;
	}

	// one hex (1 characters, 0..15)
	for (i = 0; i < 16; i++)
	{
		oneHexDrawRoutines[i] = (uint32_t)&drawCodeArea[a];

		ch1 = HexTable[i] - ' ';
		for (y = 0; y < FONT_H; y++)
		{
			const uint8_t data = fontData[(ch1*6) + y];

			index = y*(SCREEN_W / 8);
			if (index < 128)
			{
				// mov byte ptr [edi+XXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x47;
				drawCodeArea[a+2] = (uint8_t)index;
				drawCodeArea[a+3] = data;
				a += 4;
			}
			else
			{
				// mov byte ptr [edi+XXXXXXXXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x87;
				drawCodeArea[a+2] = index & 0xFF;
				drawCodeArea[a+3] = index >> 8;
				drawCodeArea[a+4] = 0x00;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = data;
				a += 7;
			}
		}

		drawCodeArea[a] = 0xC3; // retn
		a++;
	}

	// big one hex (1 characters, 0..15)
	for (i = 0; i < 16; i++)
	{
		bigOneHexDrawRoutines[i] = (uint32_t)&drawCodeArea[a];

		ch1 = HexTable[i] - ' ';
		for (y = 0; y < FONT_H; y++)
		{
			const uint8_t data = fontData[(ch1*6) + y];

			index = (y<<1)*(SCREEN_W / 8);
			if (index < 128)
			{
				// mov byte ptr [edi+XXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x47;
				drawCodeArea[a+2] = (uint8_t)index;
				drawCodeArea[a+3] = data;
				a += 4;
			}
			else
			{
				// mov byte ptr [edi+XXXXXXXXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x87;
				drawCodeArea[a+2] = index & 0xFF;
				drawCodeArea[a+3] = index >> 8;
				drawCodeArea[a+4] = 0x00;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = data;
				a += 7;
			}

			index = ((y<<1)+1)*(SCREEN_W / 8);
			if (index < 128)
			{
				// mov byte ptr [edi+XXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x47;
				drawCodeArea[a+2] = (uint8_t)index;
				drawCodeArea[a+3] = data;
				a += 4;
			}
			else
			{
				// mov byte ptr [edi+XXXXXXXXh],YYh
				drawCodeArea[a+0] = 0xC6;
				drawCodeArea[a+1] = 0x87;
				drawCodeArea[a+2] = index & 0xFF;
				drawCodeArea[a+3] = index >> 8;
				drawCodeArea[a+4] = 0x00;
				drawCodeArea[a+5] = 0x00;
				drawCodeArea[a+6] = data;
				a += 7;
			}
		}

		drawCodeArea[a] = 0xC3; // retn
		a++;
	}

	//sprintf(endMsg, "DEBUG: Pattern drawing code size = %d bytes...", a);
	return true;
}

void DrawCode_FreeCode(void)
{
	if (oldRowsOrigPtr != NULL)
	{
		free(oldRowsOrigPtr);
		oldRowsOrigPtr = NULL;
	}

	if (drawCodeAreaOrigPtr != NULL)
	{
		free(drawCodeAreaOrigPtr);
		drawCodeAreaOrigPtr = NULL;
	}
}
