; Mouse cursor routines for PT4DOS.
;
; NOTE: These are meant to only be used on a dedicated bitplane.
;       In this case, it's bitplane 4.

.386p
.MODEL FLAT

	INCLUDE "../hardware/avideo.i"

POINTER_W	EQU 8
POINTER_H	EQU 11

; mouse_t struct
m_x		EQU 0
m_y		EQU 2
m_leftbtn	EQU 4
m_rightbtn	EQU 5
m_clickhandled	EQU 6

.CODE
	EXTERN __GETDS:PROC

	; main.c
	EXTERN _mouseStructPtr:DWORD

	PUBLIC _Mouse_Interrupt
	PUBLIC _Mouse_Update

	; Main mouse interrupt
	;
	ALIGN 4
_Mouse_Interrupt PROC FAR
	pushad
	push ds
	call __GETDS
	; -----------------------------
	call _Mouse_Update
	; -----------------------------
	; reset external mouse position (so it never clamps externally)
	; -----------------------------
	mov ax,4
	mov cx,64
	mov dx,64
	int 33h
	; -----------------------------
	pop ds
	popad
	iretd
_Mouse_Interrupt ENDP

	; This routine is called both on vblank and on mouse interaction.
	;
	ALIGN 4
_Mouse_Update PROC NEAR
	pushad
	mov al,1
	xchg al,updateOngoing
	test al,al
	jnz MUEnd ; this routine is already being called!
   	; -----------------------------
   	; Set mouse button states
   	; -----------------------------
   	mov esi,_mouseStructPtr
   	; -----------------------------
	mov ax,3
	int 33h
	bt bx,0
	setc m_leftbtn[esi]
	bt bx,1
	setc m_rightbtn[esi]
	; -----------------------------
	; Read mouse position
	; -----------------------------	
	mov ax,11
	int 33h
	movsx ebx,cx
	movsx ecx,dx
	movzx eax,word ptr m_x[esi]
	movzx edx,word ptr m_y[esi]
	add ebx,eax
	add ecx,edx
	; -----------------------------
	; Clamp mouse position to screen space
	; -----------------------------
	mov eax,SCREEN_W-1
	mov edx,SCREEN_H-1
	test ebx,ebx
	jge XOK1
	xor ebx,ebx
XOK1:	cmp ebx,eax
	jle XOK2
	mov ebx,eax
XOK2:	test ecx,ecx
	jge YOK1
	xor ecx,ecx
YOK1:	cmp ecx,edx
	jle YOK2
	mov ecx,edx
YOK2:	; -----------------------------
	mov m_x[esi],bx
	mov m_y[esi],cx
	; -----------------------------
	cmp bx,oldX
	jne DrawMouse
	cmp cx,oldY
	je NoDraw
DrawMouse:
	; -----------------------------
	; Backup VGA registers
	; -----------------------------
	mov dx,3C4h
	mov ax,0202h
	out dx,al
	inc dx
	in al,dx
	xchg al,ah
	mov oldPlaneMask,ax
	; -----------------------------
	mov dx,3CEh
	mov ax,0505h
	out dx,al
	inc dx
	in al,dx
	dec dx
	xchg al,ah
	mov oldWriteMode,ax
	; -----------------------------
	mov ax,0808h
	out dx,al
	inc dx
	in al,dx
	dec dx
	xchg al,ah
	mov oldBitMask,ax	
	; -----------------------------
	mov ax,0303h
	out dx,al
	inc dx
	in al,dx
	dec dx
	xchg al,ah
	mov oldLatchMode,ax
	; -----------------------------
	; Set new VGA registers
	; -----------------------------
	mov dx,3C4h
	mov ax,0802h
	out dx,ax
	; -----------------------------
	mov dx,3CEh
	mov ax,0005h
	out dx,ax
	; -----------------------------
	mov ax,0FF0h
	out dx,ax
	; -----------------------------
	mov ax,0003h
	out dx,ax
	; -----------------------------
	; Draw mouse pointer (ebx = x, ecx = y)
	; -----------------------------
	cmp byte ptr wasDrawn,0
	je NoErase
	call _Mouse_ErasePointer
NoErase:
	call _Mouse_DrawPointer
	mov byte ptr wasDrawn,1	
	; -----------------------------
	; Restore VGA registers
	; -----------------------------
	mov dx,3C4h
	mov ax,oldPlaneMask
	out dx,ax
	; -----------------------------
	mov dx,3CEh
	mov ax,oldWriteMode
	out dx,ax
	; -----------------------------	
	mov ax,oldBitMask
	out dx,ax
	; -----------------------------	
	mov ax,oldLatchMode
	out dx,ax
	; -----------------------------
NoDraw:	mov oldX,bx
	mov oldY,cx
	mov byte ptr updateOngoing,0
	; -----------------------------
	cmp byte ptr m_leftbtn[esi],1
	je MUEnd
	cmp byte ptr m_rightbtn[esi],1
	je MUEnd
	mov byte ptr m_clickhandled[esi],0
MUEnd:	; -----------------------------
	popad
	retn
_Mouse_Update ENDP

	ALIGN 4
_Mouse_DrawPointer PROC NEAR
	push ecx
	; -------------------------------
	mov eax,ebx
	mov edx,ecx
	; -------------------------------
	mov ecx,eax
	and cl,8-1
	add ecx,8				; cl = 8+(x % 8) (block shift)
	; -------------------------------
	; setup frame buffer pointer
	shl edx,3
	lea edx,[edx+edx*4]			; edx *= 40 (SCREEN_W/8)
	shr eax,3 				; eax = x / 8
	lea edi,[VGA_FRAMEBUFFER+eax+edx]
	; -------------------------------
	lea esi,PointerData
	mov ebp,(SCREEN_W/8)*4
	; -------------------------------
	cld
	; -------------------------------
	; set right-edge clipping mask
	mov dx,00FFh
	cmp eax,(SCREEN_W/8)-1
	sbb dh,0				; if cursor outside: bx=00FF, else bx=FFFF
	jnc Clipped
	; -------------------------------
NonClipped:
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*0)],ax
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*1)],ax
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*2)],ax
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*0)],ax
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*1)],ax
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*2)],ax
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*0)],ax
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*1)],ax
	lodsw
	ror ax,cl
	mov [edi+((SCREEN_W/8)*2)],ax
	; -------------------------------
	pop ecx
	retn
	
Clipped:
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*0)],ax
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*1)],ax
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*2)],ax
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*0)],ax
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*1)],ax
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*2)],ax
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*0)],ax
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*1)],ax
	lodsw
	ror ax,cl
	and eax,edx
	mov [edi+((SCREEN_W/8)*2)],ax
	; -------------------------------
	pop ecx
	retn

_Mouse_DrawPointer ENDP

	ALIGN 4
_Mouse_ErasePointer PROC NEAR
	movzx eax,word ptr oldX
	movzx edx,word ptr oldY
	; -------------------------------
	; setup frame buffer pointer
	shl edx,3
	lea edx,[edx+edx*4]			; edx *= 40 (SCREEN_W/8)
	shr eax,3 				; eax /= 8
	lea edi,[VGA_FRAMEBUFFER+eax+edx]
	; -------------------------------
	mov edx,(SCREEN_W/8)*4
	xor eax,eax
	; -------------------------------
	mov [edi+((SCREEN_W/8)*0)],ax
	mov [edi+((SCREEN_W/8)*1)],ax
	mov [edi+((SCREEN_W/8)*2)],ax
	mov [edi+((SCREEN_W/8)*3)],ax
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)],ax
	mov [edi+((SCREEN_W/8)*1)],ax
	mov [edi+((SCREEN_W/8)*2)],ax
	mov [edi+((SCREEN_W/8)*3)],ax
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)],ax
	mov [edi+((SCREEN_W/8)*1)],ax
	mov [edi+((SCREEN_W/8)*2)],ax
	retn
_Mouse_ErasePointer ENDP

	ALIGN 4
PointerData:
	dw 1000000000000000b
	dw 1100000000000000b
	dw 1110000000000000b
	dw 1111000000000000b
	dw 1111100000000000b
	dw 1111110000000000b
	dw 1111111000000000b
	dw 1111111100000000b
	dw 1111100000000000b
	dw 1101110000000000b
	dw 1000110000000000b
	
	ALIGN 2
oldPlaneMask	dw 0
oldWriteMode	dw 0
oldBitMask	dw 0
oldLatchMode	dw 0
oldX		dw 8000
oldY		dw 8000
updateOngoing	db 0
wasDrawn	db 0

END
