// some of this code is partly based on Fasttracker II's code

#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include "pit.h"
#include "irq.h"
#include "video.h"
#include "../audio/audio.h"

#define TIMER_IRQ 0

// globals
uint32_t dmaMixPos, oldInt8VectorDelta, oldInt8VectorFrac;
uint32_t samplesPerFrame15; // * 1.5
uint32_t samplesPerFrame25; // * 2.5
void __interrupt (*oldInt8Vector)(void);

// apitint.asm
extern void __interrupt int8Vector(void);
extern void __interrupt int8VectorNoAudio(void);

void PIT_StartVblankInterrupt(void)
{
	double dOldTimerHz = PIT_CLK / 65536.0; // ~18.2065Hz (clock for original int0 on PIT channel 0)
	double dNewTimerHz = PIT_CLK / oneFrameTime;

	oldInt8Vector = IRQ_GetIntVector(TIMER_IRQ);
	oldInt8VectorDelta = (uint32_t)(((dOldTimerHz / dNewTimerHz) * (UINT32_MAX+1.0)) + 0.5); // rounded 0.32fp
	oldInt8VectorFrac = 0;

	if (audioOpen)
	{
		const double dSamplesPerFrame = audio.dOutputFreq / dNewTimerHz;
		const uint32_t samplesPerFrame = (uint32_t)(dSamplesPerFrame + 0.5);

		samplesPerFrame15 = samplesPerFrame + (samplesPerFrame >> 1);
		samplesPerFrame25 = (samplesPerFrame << 1) + (samplesPerFrame >> 1);
		dmaMixPos = (3 * (samplesPerFrame << audio.sampleShift)) & ~31;

		IRQ_SetIntVector(TIMER_IRQ, int8Vector);
	}
	else
	{
		// no audio
		IRQ_SetIntVector(TIMER_IRQ, int8VectorNoAudio);
	}

	// set PIT channel 0 timer
	_disable();
	outp(0x43, 0x36); // channel 0, mode 3
	outp(0x40, oneFrameTime & 0xFF);
	outp(0x40, oneFrameTime >> 8);
	Video_WaitVBL();
	IRQ_Enable(TIMER_IRQ);
	_enable();

	vertIntOn = true;
}

void PIT_StopVblankInterrupt(void)
{
	if (vertIntOn) 
	{
		// restore PIT channel 0 back to default (BIOS timer)
		_disable();
		outp(0x43, 0x36); // channel 0, mode 3
		outp(0x40, 0); // ~18.2065Hz (0 = period 65536)
		outp(0x40, 0);
		_enable();

		IRQ_RestoreIntVector(TIMER_IRQ);
		vertIntOn = false;
	}
}
